#!/usr/bin/perl

# Initially created to follow an active broadsoft billing CDR file and generate
# radius accounting records from it on the wire to a Radiator radius server. 
# Could easily be adapted to other sources and kinds of CDR records. 
#
# Supports a primary and backup radius server with auto fallback to the primary
# when possible. 
# 
# It expects your radius server to do some form of de-duping of records. If your
# billing file doesn't get any activity for a while the file will be rescanned 
# and records retransmitted as part of the normal process of looking for a rotated
# billing file. 
#
# 
# By: Jared Watkins jared at jaredwatkins dot com
# © Copyright Jared Watkins 2012
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#            
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

use File::Tail;
use Time::HiRes qw( ualarm );
use Net::Inet qw(:routines);
use Net::Radius::Dictionary;
use Net::Radius::Packet;
use Net::Gen qw(:af);
use POSIX qw(uname setsid);
use Net::UDP;
use Carp;
#use warnings;
use Getopt::Std;
use strict;
use Sys::Syslog;
use Digest::MD5 qw/md5_hex/;


my %o = ();
my ($bwactivedir, $senderip);
getopts("d:f:a:t", \%o);
$Getopt::Std::STANDARD_HELP_VERSION = 1;
my @archivefiles;

#### CONFIG #################################
my $radserver       = '192.168.103.201';
my $radserverbackup = '192.168.103.200';
my $radsecret       = 'test123';
my $acctport        = '1813';
#############################################

# Directory where the active billing file is located
if ($o{d})
{
    $bwactivedir = $o{d};
} else {
    $bwactivedir     = '/bw/broadworks/billing/active';
}

# Archive mode... for loading old data
# Generate a list of files to be loaded... this could be smarter or more flexible
if (defined($o{a}))
{
    if (not $o{d}) { $o{d} = "/bw/broadworks/billing/archive"; }

    my @lines = `find "$o{d}/" -mtime +$o{a} -iname "*.csv"`;
    foreach my $line (@lines)
    {
        trim($line);
        push @archivefiles, $line;
    }
}

# F for fake ip
if ($o{f})
{
    $senderip = $o{f};
} else {
    # discover my own IP address
    $senderip = join '.',unpack "C4",gethostbyname((uname)[1]);
}

my $dict = Net::Radius::Dictionary->new( './dictionary', 
                                         './dictionary.broadsoft',
                                       ) or croak "No dictionary found\n";

###############################

daemonize();



# Init everything... 
openlog('file2radius', 'pid,nofatal', 'user');

# Create the network socket
my $s = new Net::UDP or croak $!;

my ($acctaddr, $paddr,$acctaddr_backup, $paddr_backup, $activeserver, $activeip);
$paddr = gethostbyname($radserver) or croak "Can't resolve host $radserver\n";
$acctaddr = pack_sockaddr_in(AF_INET, $acctport, $paddr);

$paddr_backup = gethostbyname($radserverbackup) or croak "Can't resolve host $radserverbackup\n";
$acctaddr_backup = pack_sockaddr_in(AF_INET, $acctport, $paddr_backup);


my $filename = newfilename();

my $file = File::Tail->new(name=>$filename, maxinterval=>5, interval=>1, tail=>-1,name_changes=>\&newfilename);


if (defined($o{a}))
{
    syslog('info', "Archive run starting up on directory $o{d}");
} else { syslog('info', "Starting up.. tailing file $filename"); }



sub newfilename
{
    my $file;
    
    $file = `ls -tr1 $bwactivedir/*.csv |tail -n 1`;
    # Every time the file rolls.. try to switch back to the primary server
    # Also called when a file has been idle for a while
    if ($activeip ne $radserver)
    {
        $activeserver = $acctaddr;
        $activeip     = $radserver;
        syslog('info', "Switching back to primary radius server on new cdr file");
    }

    return trim($file);
}
        
sub trim($)
{
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

my ($cdrversion, $whence, $sessionid);
my $ident = 0;

my $fp;
FILE: if (defined($o{a}))
      {
          my $arfile = trim(shift @archivefiles);
          if (not $arfile) { syslog('info', "Archive run complete"); exit 0;}
          syslog('info', "Archive Run in Progress. Loading file: $arfile");
          
          open($fp, '<', "$arfile") or syslog('info', "Unable to open archive $! file:$arfile:");
      }
    

LINE: while (1)
{
    my $line;
    
    if (defined($o{a}))
    {
        $line = trim(<$fp>);
        if (not $line) { close $fp; goto FILE; }
        
    } else {
        $line = trim($file->read);
    }
    
    my ($rec, $req, $resp);
    my %a = ();   # Radius standard attribs
    my %va = ();  # Vendor attribs
    
    my @d = split /,/, $line;

    # Version of the cdr format is only on the first line of the file.. ie 16 18 etc 
    if ($d[0] =~ /^version=(\d+\.\d+)\s/)
    {
        $cdrversion = $1;
        next;
    }

    if ($d[2] and $d[2] =~ /start/i) { next; }
    if ($d[2] and $d[2] =~ /end/i)   { next; } # End of cdr file
    
    #print "$line\n";
    unshift (@d, '');  # Shift by one so positions line up with the dictionary for the for loop below
    
    $va{'BroadSoft'}{$dict->vsattr_name(6431, 0)} = $cdrversion;
    
    # Create an auth-request packet
    $req = new Net::Radius::Packet $dict;
    
    $req->set_attr('NAS-IP-Address' => $senderip);
    $req->set_code('Accounting-Request');


    
    # Loop through the cdr fields and store them in the packet if they are defined
    # pulling the radius attrib names from the broadsoft dictionary
    # Hash within a hash in case you need to set more than one vendor id in the same packet (valid?)
    #
    for (my $j=0; $j <= $#d; $j++)
    {
        #if ($dict->vsattr_has_val(6431, $j))
        #{
            my $field = $dict->vsattr_name(6431, $j);
            #print "$j | $field | $d[$j]\n";
            if ($d[$j]) { $va{'BroadSoft'}{$field} = $d[$j]; } else { next; }

        #} else { next; } # Skip it if it's not in the dictionary
    }
    

    my $type;
    
    if (not $va{'BroadSoft'}{'BWAS-Direction'}) { next; }
    
    # Skip terminating recs if asked to.. 
    if (defined($o{t}) and $va{'BroadSoft'}{'BWAS-Direction'} eq 'Terminating') { next; }

    my $auth = substr(md5_hex($$ . rand(4096)), 0, 16);
    $req->set_authenticator($auth);
    $ident = $ident & 255;
    $req->set_identifier($ident);

    
    # Figure out what kind of record this is (start|stop|interum).. not explicitly defined in the CDR entry
    # You have to figire it out from the presense or absense of things
    # 
    if ($va{'BroadSoft'}{'BWAS-Type'} eq 'Normal' 
        and not $va{'BroadSoft'}{'BWAS-Release-Time'} 
        and not $va{'BroadSoft'}{'BWAS-Termination-Cause'} 
        and not $va{'BroadSoft'}{'BWAS-Releasing-Party'} ) 
    { $type = 'Start'; }
    elsif ($va{'BroadSoft'}{'BWAS-Type'} eq 'Normal' 
           and $va{'BroadSoft'}{'BWAS-Release-Time'} 
           and $va{'BroadSoft'}{'BWAS-Termination-Cause'} 
           and $va{'BroadSoft'}{'BWAS-Releasing-Party'} ) 
    { $type = 'Stop'; }
    elsif ($va{'BroadSoft'}{'BWAS-Type'} eq 'Long Duration')
    { $type = 'Alive'; }
    
    $req->set_attr('Acct-Status-Type', $type);
    
    # Populate the 'standard' radius attribs
    foreach my $key (sort keys %a)
    {
        $req->set_attr($key, $a{$key});
    }
    
    # Populate the vendor specific attribs.. walking our hash within a hash structure
    foreach my $vendorid (sort keys %va) 
    {
        foreach my $key (sort keys %{$va{$vendorid}} )
        {
            $req->set_vsattr($vendorid, $key, $va{$vendorid}{$key});
        }
    }
    
    # This mimics the way broadsoft generates a radius session id which is unrelated to
    # BW session id.  Must be unique to the call.. but common to the recs that are 
    # pieces of a call.. ie Start, Alive, Stop must share a sessionid for the same call
    #
    $sessionid = genSessionIdBW($va{'BroadSoft'}{'BWAS-Record-id'}, 
                                $va{'BroadSoft'}{'BWAS-Start-Time'},
                                $va{'BroadSoft'}{'BWAS-Local-Callid'},
                                $type);

    $req->set_attr('Acct-Session-Id', $sessionid); 


    # TODO Figure out a queue for stuff we can't send?
    my $count = 3;
    my $secondchance;
    TRY: while ($count)
    {
        eval
        {
            local $SIG{ALRM} = sub { die 'foo' };

            ualarm(1_500_000);

            # Calls a method of Net::UDP to actually send the packet
            $s->sendto($req->pack, $activeserver);
            #$req->dump;

            # wait for response
            #print "DEBUG Waiting for reply\n";
            $rec = $s->recv(undef, undef, $whence);

            # This gets the reply packet
            $resp = new Net::Radius::Packet $dict, $rec;

            #$resp->dump;
            ualarm(0);

        };

        if ($@)  # Timed out.. should contain the string foo from the above error handler
        {
                syslog('warning', "Didn't get a radius reply count: $count!\n");
                --$count;
                # If we have already looped 3 times.. flip over to the other radius 
                # server and try it 3 times
                #
                if (not $count and not $secondchance)
                {
                    $secondchance = 1;
                    $count = 3;

                    if ($activeip eq $radserver)  # Can't use the packed UDP connection object for comparisons.. 
                    {
                        syslog('warning', "WARN: Switching to backup radius server");

                        $activeserver = $acctaddr_backup;
                        $activeip     = $radserverbackup;
                    } else {
                        syslog('warning', "WARN: Switching to primary radius server");

                        $activeserver = $acctaddr;
                        $activeip     = $radserver;
                    }
                }
                next TRY;

        } else {
        
            ++$ident;
            next LINE;
        }
    }
}

#######################

my %sessionids;
sub genSessionIdBW
{
    my ($raw, $starttime, $localcallid, $type) = @_;
    my $systemid;
    
    # First part of the $raw.. padded with 0s to 10 characters - Don't actually use for radius session id
    my $eventcounter = substr $raw, 0, 10;
    

    # Pull out the yearmonday part of the date to test against the raw systemid coming in
    # I do this to determine how many characters long the BW system id is 
    # Possible lengths are 8, 12, or 36 characters!
    my $testdate = substr $starttime, 0, 8;
    my $sesdate  = substr $starttime, 0, 18;

    # First test for 8 char length
    my $tmp = substr $raw, 18, 8;
    if ($testdate eq $tmp)
    {
        # using an 8 character systemid
        $systemid = substr $raw, 10, 8;
        
    } else {
        $tmp = substr $raw, 22, 8;
        if ($testdate eq $tmp)
        {
            # using a 12 character systemid
            
            $systemid = substr $raw, 10, 12;

        } else {
            $tmp = substr $raw, 46, 8;
            if ($testdate eq $tmp)
            {
                # using a 36 character systemid
                $systemid = substr $raw, 10, 36;
            }
        }
    }
    
    my $sessionid = $localcallid . $systemid . $sesdate;

    if ($type =~ /start/i or $type =~ /alive/i or $type =~ /interim/i)
    {
        $sessionids{$sessionid} = 1;
        return $sessionid;

    } elsif ($type =~ /stop/i) {
        
        delete $sessionids{$sessionid};
        return $sessionid;
    }
}

sub daemonize
{
    #chdir '/'               or croak "Can't chdir to /: $!";
    open STDIN, '/dev/null'   or croak "Can't read /dev/null: $!";
    open STDOUT, '>/dev/null' or croak "Can't write to /dev/null: $!";
    defined(my $pid = fork) or croak "Can't fork: $!";
    exit if $pid;
    croak "Can't start a new session: $!" if setsid == -1;
    open STDERR, '>&STDOUT' or croak "Can't dup stdout: $!";
}
 

sub HELP_MESSAGE
{
  print "
Syntax

$0 <-d directory> <-f ip address>

    -d [directory path]      Optional full path to the location of the CSV file you want to tail
                             Defaults to /bw/broadsoft/billing/active

    -f [ip address]          Used when 'faking' a source ip. Useful when testing or developing 
                             by tailing a local file and making it appear to come from a remote
                             source by inserting this IP into the NAS-IP-Address radius attrib

    -a [days]                Run in 'archive mode' - Scan the indicatd directory and send 
                             (not tail) all .csv files that were modified x days ago (based on file mtime)
                             So to send all files use 0.. files older than 7 days would be '7' etc. 

    -t                       Silently ignore Terminating records.. do not send to the radius server 
";
  exit;
}
                                                        